// Single-sum checksum examples -- Chapter 5 -- unit tests

// Copyright 2024 Philip Koopman
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//    http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "support.h"
#include "chapter5_examples.h"

static const uint32_t initialSeed = SEED;  // Initial sum

// Test generic modular addition
uint32_t TestAdd8(uint8_t dataWord[], uint32_t dwSize, uint32_t modulus)
{ 
  uint32_t resultA = (uint32_t) Add8_A(dataWord, dwSize, modulus);
  uint32_t resultB = (uint32_t) Add8_B(dataWord, dwSize, modulus);
  uint32_t resultC = (uint32_t) Add8_C(dataWord, dwSize, modulus);
  uint32_t resultD = (uint32_t) Add8_D(dataWord, dwSize, modulus);
  printf("Modulus=0x%02X\n", modulus);
  printf("Result Add8_A=    0x%02X\n", resultA);
  printf("Result Add8_B=    0x%02X\n", resultB);
  printf("Result Add8_C=    0x%02X\n", resultC);
  printf("Result Add8_D=    0x%02X\n", resultD);

  // Make sure they all got the same result  
  assert(resultA == resultB);
  assert(resultB == resultC);
  assert(resultC == resultD);

  return(resultA);
}

// Test LRC
uint32_t TestLRC(uint8_t dataWord[], uint32_t dwSize)
{ uint32_t modulus = 256;
  uint32_t resultX = (uint32_t) LRC(dataWord, dwSize, modulus);
  printf("Result LRC=\t0x%02X\n", resultX);
  return(resultX);  
}

// Test two's complement modular addition
uint32_t TestTwos(uint8_t dataWord[], uint32_t dwSize)
{ uint32_t modulus = 256;
  printf("Two's Complement;\t");
  uint32_t resultA = TestAdd8(dataWord, dwSize, modulus);
  uint32_t resultE = (uint32_t) Add8_Twos(dataWord, dwSize, modulus);
  printf("Result Add8_Twos= 0x%02X\n", resultE);

  // Make sure they all got the same result
  assert(resultA == resultE);
  return(resultA);
}

// Test ones' complement modular addition
uint32_t TestOnes(uint8_t dataWord[], uint32_t dwSize)
{ uint32_t modulus = 255;
  printf("ones' Complement;\t");
  uint32_t resultA = TestAdd8(dataWord, dwSize, modulus);
  uint32_t resultF = (uint32_t) Add8_Ones(dataWord, dwSize, modulus);
  printf("Result Add8_Ones= 0x%02X\n", resultF);
  
  // convert -0 to +0 for comparison
  if (resultF == 0xFF) { resultF = 0x00; }

  // Make sure they all got the same result
  assert(resultA == resultF);
  return(resultA);
}

// Test modulus 253
uint32_t TestOther(uint8_t dataWord[], uint32_t dwSize)
{ uint32_t modulus = 253;
  uint32_t resultA = TestAdd8(dataWord, dwSize, modulus);
  return(resultA);  
}

// -------------------------------
// Run all unit tests
int main(void) 
{ 
#ifdef VALIDATE
  const uint32_t dwAlloc = 16; // divisible by 4 and greater than 9
  uint32_t dwSize = 9; 
#else // Random consistency check among implementations
  const uint32_t dwAlloc = 1200; // divisible by 4 and greater than 9
  uint32_t dwSize = 1200; 
#endif

  uint8_t dataWord[dwAlloc];
  InitializeData(dataWord, dwAlloc); // initialize padding bytes too
  PrintWord(dataWord, dwSize);

  printf(">>>>> Initial Seed = %d\n", initialSeed);

  uint32_t result = 0;
  
  result = TestLRC(dataWord, dwSize);
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0x31); } else 
  if(initialSeed == 0x9D) { assert(result == 0xAC); } else 
  if(initialSeed == 0xFF) { assert(result == 0xCE); } else 
  { assert(false); }  
#endif  

  result = TestTwos(dataWord, dwSize);
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0xDD); } else 
  if(initialSeed == 0x9D) { assert(result == 0x7A); } else 
  if(initialSeed == 0xFF) { assert(result == 0xDC); } else 
  { assert(false); }  
#endif  

  result = TestOnes(dataWord, dwSize);
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0xDE); } else 
  if(initialSeed == 0x9D) { assert(result == 0x7C); } else 
  if(initialSeed == 0xFF) { assert(result == 0xDE); } else 
  { assert(false); }  
#endif  

  result = TestOther(dataWord, dwSize);
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0xE0); } else 
  if(initialSeed == 0x9D) { assert(result == 0x80); } else 
  if(initialSeed == 0xFF) { assert(result == 0xE2); } else 
  { assert(false); }  
#endif 
   
  printf("<<<<<<<<<<<<<<<<<<<<<<<<<<< PASS >>>>>>>>>>>>>>>>>>>>>>>>>>>\n");   
  result = EXIT_SUCCESS;
  return((int)result) ;
}
